#ifndef _BUD_H
#define _BUD_H

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include "list.h"

#define MAX_INCLUDES (8)

#define LOG_LEVEL_DEBUG (0)
#define LOG_LEVEL_INFO  (1)
#define LOG_LEVEL_WARN  (2)
#define LOG_LEVEL_ERR   (3)

#define gettid() syscall(__NR_gettid)

typedef enum tagPringLanguage{
    PL_ESC_P            = 0,
    PL_ESC_P_2,
    PL_EPSON_SCRIPT,
    PL_PCL_5,
    PL_PCL_XL,
    PL_UFR,
    PL_UFR_II,
    PL_BR_SCRIPT_2,
    PL_BR_SCRIPT_3,
    PL_UNKNOWN,
}PrintLanguage;

typedef struct tagBlockInfo{
    unsigned int offset;/*first byte offset to the file head*/
    unsigned int count; /*element count*/
}BlockInfo;

typedef struct tagStringTable{
    char* buf;
    int pos;
    int size;
}StringTable;

typedef struct tagBudFileHeaderV1{
    char id[8];             /*budfake*/
    char author[8];
    char version[8];
    /*the number of seconds since the Epoch, 1970-01-01 00:00:00 +0000 (UTC)*/
    unsigned int time_stamp;
    unsigned int file_size;
    char print_language[16]; /*ESC/P, ESC/P 2, PCL 5, PCL XL, UFR, UFR II, etc...*/
    char color_space[6];    /*w(white),k(black),rgb,cmy,cmyk*/;
    unsigned char bits_per_pixel;
    unsigned char sub_blocks;
    BlockInfo attributes;  /*root level*/
    BlockInfo features;
    BlockInfo commands;    /*root level*/
    BlockInfo switch_cases;/*root level*/
    BlockInfo string_table;
}BudFileHeaderV1;

struct tagRect{
    unsigned int left;
    unsigned int top;
    unsigned int right;
    unsigned int bottom;
};

struct tagPair{
    unsigned int first;
    unsigned int second;
};

struct tagStrList{
    char type; /*c*/
    char** table;
    unsigned int count;
};

struct tagNumList{
    char type; /*K*/
    unsigned int* table;
    unsigned int count;
};

struct tagAST{
    char nodetype; /*+,-,*,/,%,m,M,N*/
    struct tagAST* left;
    struct tagAST* right;
};

struct tagNumNode{
    char nodetype; /*type K, no child node*/
    unsigned int number;
};

struct tagVarNode{
    char nodetype; /*type V, no child node*/
    char* name;
};

struct tagCmdArg{
    char type; /*format type:c,l,d, etc*/
    struct tagAST* ast;
    unsigned int min;
    unsigned int max;
};

struct tagArgList{
    char count;
    struct tagCmdArg* table;
};

struct tagEntryList {
    char            type; /*'a', 'e', 'p', 'h'*/
};

typedef enum tagGpdType{
    GPD_NUMBER = 0,
    GPD_STRING,
    GPD_PAIR,
    GPD_RECT,
    GPD_LIST,
    GPD_EXPR,
    GPD_MACRO,
    GPD_MACRO_REF,
}GpdType;

struct tagAttribute{
    struct tagEntryList     node;
    GpdType                 type;
    char*                   name;
    union {
        unsigned int        num;
        char*               str;
        struct tagPair      pair;
        struct tagRect      rect;
        struct tagNumList   ilist;
        struct tagStrList   slist;
        struct tagArgList   arg;
    };
};

typedef enum tagPreproCmd{
    GPD_PREP_INCLUDE = 0,
    GPD_PREP_DEFINE,
    GPD_PREP_UNDEFINE,
    GPD_PREP_IFDEF,
    GPD_PREP_ELSEIFDEF,
    GPD_PREP_ELSE,
    GPD_PREP_ENDIF,
}PreproCmd;

struct tagPrepro{
    struct tagEntryList     node;
    PreproCmd               cmd;
    char*                   value;
};

typedef enum tagEntryType{
    ENTRY_FEATURE = 0,
    ENTRY_OPTION,
    ENTRY_COMMAND,
    ENTRY_SWITCH,
    ENTRY_CASE,
    ENTRY_IGNOREBLOCK,
    ENTRY_FONTCAR,
}EntryType;

typedef enum tagCommandKind{
    CMD_JOB_SETUP = 0,
    CMD_DOC_SETUP,
    CMD_PAGE_SETUP,
    CMD_PAGE_FINISH,
    CMD_DOC_FINISH,
    CMD_JOB_FINISH,
    CMD_CURSOR_MOVE,
    CMD_RASTER_SEND,
    CMD_GENERIC_SET, /*text style, font select, color select*/
    CMD_KIND_COUNT,
}CommandKind;

#define GET_CMD_KIND(v) ((v)>>24)
#define GET_CMD_ORDER(v) ((v)&0x00ffffff)

struct tagEntry{
    struct tagEntryList     node;
    EntryType               type;
    char*                   name;
    struct tagEntryList**   sub;     /* can be NULL for simple entry*/
    int                     count;   /*number of sub entry*/
    struct tagEntryList*    parent;  /* can be NULL for root entry*/

    /*bit 24-31, command kind: cursor, text, font, setup, etc...
      bit  0-23, command execute order.
    */
    unsigned int            cmd_extral;
};

struct tagCommandList{
    struct list          entry;
    struct tagEntry*     cmd;
};

#endif /*end #ifndef _BUD_H*/

