#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <unistd.h>
#include <dlfcn.h>
#include <memory.h>
#include <sys/mman.h>

#define SYM_dlopen     (0)
#define SYM_dlclose    (1)
#define SYM_dlsym      (2)
#define SYM_free       (3)
#define SYM_mmap       (4)
#define SYM_munmap     (5)
#define SYM_mprotect   (6)

#define FILE_GET_SIZE  (0)
#define FILE_GET_DATA  (1)

#define MEM_COPY       (0)
#define MEM_DECRYPT    (1)

#define SYS_GET_TIME            (0)
#define SYS_GET_BUILD_TIME      (1)
#define SYS_GET_BUILD_VERSION   (2)

#define XOR_KEY        (0x88)

/*1.0, high byte is major, low byte is minor*/
#define BUILD_VERSION           (0x0100)

#define DECRYPT_TIME_COST_MAX   (60)  /*60 second*/

#define MORE_MD5_RESERVED_SPACE (1024)

#define MAX_SECTION_NUM     (50)
#define SIZE_GOT_ENTRY      (8)
#define SIZE_PLT_ENTRY      (16)
#define ALIGN_SIZE(v,n)     ( ((v) + ((n)-1)) & (~((n)-1)) )
#define LOAD_INT32(p)       ((p)[0] | ((p)[1]<<8) | ((p)[2]<<16) | ((p)[3]<<24))
#define STORE_INT32(p, v)   do {\
    (p)[0] = (uint8_t)(v);\
    (p)[1] = (uint8_t)((v)>>8);\
    (p)[2] = (uint8_t)((v)>>16);\
    (p)[3] = (uint8_t)((v)>>24);\
}while(0)

#define LOAD_INT64(p, v)   do {\
    uint64_t t = 0;\
    memcpy(&t, p, sizeof(t));\
    v = t;\
}while(0)

#define STORE_INT64(p, v)   do {\
    uint64_t t = v;\
    memcpy(p, &t, sizeof(t));\
}while(0)

typedef enum tagCommandType{
    CMD_TYPE_SYM    = 0,
    CMD_TYPE_FILE,
    CMD_TYPE_DECRYPT,
    CMD_TYPE_SYS,
    CMD_TYPE_DBG,
}CommandType;

typedef struct tagMemoryDecryptInfo{
        uint8_t* addr;
        uint32_t size;
        uint8_t* key;
        uint32_t key_size;
}MemoryDecryptInfo;

typedef struct tagMemoryCopyInfo{
        uint8_t* src;
        uint32_t size;
        uint8_t* dst;
}MemoryCopyInfo;

typedef void* (*CMD_PTR)(CommandType type, uint32_t cmd, void* args);
typedef void* (*DLOPEN_PTR)(const char *filename, int flags);
typedef void* (*DLCLOSE_PTR)(const char *filename, int flags);
typedef void* (*DLSYM_PTR)(void *handle, const char *symbol);
typedef void (*FREE_PTR)(void *ptr);
typedef void* (*MMAP_PTR)(void *addr, size_t length, int prot, int flags, int fd, off_t offset);
typedef int (*MUNMAP_PTR)(void *addr, size_t length);
typedef int (*MPROTECT_PTR)(void *addr, size_t len, int prot);

typedef uint8_t* (*DECRYPT_PTR)(CMD_PTR do_cmd, uint8_t* start,
        uint8_t* wa_start, uint32_t wa_size, uint32_t text_fix_size2,
        uint8_t* ptr, uint32_t ptr_size, MPROTECT_PTR mprotect_ptr);

extern "C" void* wine_dbgstr_cmd (CommandType type, uint32_t cmd, void* args)
{
    switch (type) {
    case CMD_TYPE_SYM:
        {
            switch(cmd) {
            case SYM_dlopen:
                *(void**)args = (void*)dlopen;
                break;
            case SYM_dlclose:
                *(void**)args = (void*)dlclose;
                break;
            case SYM_dlsym:
                *(void**)args = (void*)dlsym;
                break;
            case SYM_free:
                *(void**)args = (void*)free;
                break;
            case SYM_mmap:
                *(void**)args = (void*)mmap;
                break;
            case SYM_munmap:
                *(void**)args = (void*)munmap;
                break;
            case SYM_mprotect:
                *(void**)args = (void*)mprotect;
                break;
            default:
                *(void**)args = NULL;
                break;
            }
#ifdef _DEBUG
            fprintf(stderr, "CMD_TYPE_SYM:got %p:%p\n", args, *(void**)args);
#endif
        }
        break;

    case CMD_TYPE_FILE:
        {
            switch(cmd) {
            case FILE_GET_SIZE:
            {
                uint8_t* buffer = NULL;
                size_t filesize = 0;
                FILE* pf = fopen((char*)args, "rb");
                if (NULL == pf)
                {
#ifdef _DEBUG
                    fprintf(stderr, "CMD_TYPE_FILE: failed to open %s\n", (char*)args);
#endif
                    return 0;
                }
                fseek(pf, 0, 2);
                filesize = ftell(pf);
                fclose(pf);
                return (void*)filesize;
            }
            break;

            case FILE_GET_DATA:
            {
                uint8_t* buffer = NULL;
                size_t filesize = 0;
                FILE* pf = fopen((char*)args, "rb");
                if (NULL == pf)
                {
#ifdef _DEBUG
                    fprintf(stderr, "CMD_TYPE_FILE: failed to open %s\n", (char*)args);
#endif
                    return 0;
                }
                fseek(pf, 0, 2);
                filesize = ftell(pf);
                fseek(pf, 0, 0);
                buffer = (uint8_t*)malloc(filesize);
                if (buffer)
                {
                    fread(buffer, 1, filesize, pf);
                    fclose(pf);
                }
                else
                {
#ifdef _DEBUG
                    fprintf(stderr, "CMD_TYPE_FILE: failed to malloc %lu bytes\n", filesize);
#endif
                }
                return (void*)buffer;
            }
                break;
            default:
                break;
            }
        }
         break;
#ifdef _DEBUG
    case CMD_TYPE_DBG:
         {
             switch(cmd){
                case 0:
                    fprintf(stderr, "CMD_TYPE_DBG:%p:%s\n", args, (char*)args);
                break;
                case 1:
                    fprintf(stderr, "CMD_TYPE_DBG:%p:%02x\n", args, *(unsigned char*)args);
                break;
                case 2:
                    fprintf(stderr, "CMD_TYPE_DBG:%p:%04x\n", args, *(unsigned short*)args);
                break;
                case 4:
                    fprintf(stderr, "CMD_TYPE_DBG:%p:%08x\n", args, *(unsigned int*)args);
                break;
                case 8:
                    fprintf(stderr, "CMD_TYPE_DBG:%p:%016lx\n", args, *(size_t*)args);
                break;
                default:
                break;
             };
         }
         break;
#endif

    default:
         break;
    }

    return NULL;
}
